@@ -39,7 +39,7 @@ class Noise(Field):
         shape = (self._shape if shape is None else shape) & resolution
         rndj = math.to_complex(random_normal(shape)) + 1j * math.to_complex(random_normal(shape))  # Note: there is no complex32
         with math.NUMPY:
-            k = math.fftfreq(resolution) * resolution / size * self.scale  # in physical units
+            k = math.fftfreq(resolution) * resolution / math.tensor(size) * math.tensor(self.scale)  # in physical units
             k = math.vec_squared(k)
         lowest_frequency = 0.1
         weight_mask = math.to_float(k > lowest_frequency)
