@@ -19,13 +19,13 @@ from .physics import fluid, flip, advect, diffuse
 
 # Classes
 from .math import DType, Solve
-from .geom import Geometry, Sphere, Box
+from .geom import Geometry, Sphere, Box, Cuboid
 from .field import Grid, CenteredGrid, StaggeredGrid, GeometryMask, SoftGeometryMask, HardGeometryMask, Noise, PointCloud, Scene
 from .vis import view, Viewer, control
 from .physics._boundaries import Obstacle
 
 # Constants
-from .math import PI
+from .math import PI, INF, NAN
 
 # Functions
 from .math import wrap, tensor, spatial, channel, batch, instance
