@@ -17,8 +17,8 @@ from .backend._dtype import DType
 from .backend import NUMPY, precision, set_global_precision, get_precision
 
 from ._config import GLOBAL_AXIS_ORDER
-from ._shape import Shape, EMPTY_SHAPE, spatial, channel, batch, instance, merge_shapes, concat_shapes
-from ._tensors import wrap, tensor, Tensor, TensorDim, TensorLike, Dict
+from ._shape import Shape, EMPTY_SHAPE, spatial, channel, batch, instance, merge_shapes, concat_shapes, IncompatibleShapes
+from ._tensors import wrap, tensor, layout, Tensor, TensorDim, TensorLike, Dict, to_dict, from_dict, shape
 from .extrapolation import Extrapolation
 from ._ops import (
     choose_backend_t as choose_backend, all_available, convert, seed,
@@ -37,8 +37,9 @@ from ._ops import (
     abs_ as abs, sign,
     round_ as round, ceil, floor,
     maximum, minimum, clip,
-    sqrt, exp, sin, cos, tan, log, log2, log10,
+    sqrt, exp, sin, cos, tan, log, log2, log10, sigmoid, arcsin, arccos,
     to_float, to_int32, to_int64, to_complex, imag, real, conjugate,
+    degrees,
     boolean_mask,
     isfinite,
     closest_grid_values, grid_sample, scatter, gather,
@@ -49,7 +50,7 @@ from ._ops import (
 )
 from ._nd import (
     shift,
-    vec_abs, vec_squared, vec_normalize, cross_product,
+    vec_abs, vec_abs as vec_length, vec_squared, vec_normalize, cross_product, rotate_vector,
     normalize_to,
     l1_loss, l2_loss, frequency_loss,
     spatial_gradient, laplace,
@@ -59,14 +60,24 @@ from ._nd import (
 )
 from ._functional import (
     LinearFunction, jit_compile_linear, jit_compile,
-    functional_gradient, custom_gradient, print_gradient,
+    functional_gradient, functional_gradient as gradient, custom_gradient, print_gradient, hessian,
     solve_linear, solve_nonlinear, minimize, Solve, SolveInfo, ConvergenceException, NotConverged, Diverged, SolveTape,
+    map_types, map_s2b,
 )
 
 
 PI = 3.14159265358979323846
 """Value of π to double precision """
-pi = PI
+pi = PI  # intentionally undocumented, use PI instead. Exists only as an anlog to numpy.pi
+
+INF = float("inf")
+""" Floating-point representation of positive infinity. """
+inf = INF  # intentionally undocumented, use INF instead. Exists only as an anlog to numpy.inf
+
+
+NAN = float("nan")
+""" Floating-point representation of NaN (not a number). """
+nan = NAN  # intentionally undocumented, use NAN instead. Exists only as an anlog to numpy.nan
 
 NUMPY = NUMPY  # to show up in pdoc
 """Default backend for NumPy arrays and SciPy objects."""
