@@ -10,8 +10,8 @@ import torch.fft
 import torch.nn.functional as torchf
 
 from phi.math import DType
-from phi.math.backend import Backend, NUMPY, ComputeDevice
-from phi.math.backend._backend import combined_dim, SolveResult
+from phi.math.backend import Backend, NUMPY, ComputeDevice, PHI_LOGGER
+from phi.math.backend._backend import combined_dim, SolveResult, get_functional_derivative_order
 
 
 class TorchBackend(Backend):
@@ -19,7 +19,8 @@ class TorchBackend(Backend):
     def __init__(self):
         cpu = NUMPY.cpu
         self.cpu = ComputeDevice(self, "CPU", 'CPU', cpu.memory, cpu.processor_count, cpu.description, ref='cpu')
-        Backend.__init__(self, 'PyTorch', default_device=self.cpu)
+        gpus = self.list_devices('GPU')
+        Backend.__init__(self, 'PyTorch', default_device=gpus[0] if gpus else cpu)
 
     def prefers_channels_last(self) -> bool:
         return False
@@ -40,8 +41,11 @@ class TorchBackend(Backend):
                                              ref=f'cuda:{index}'))
         return devices
 
+    def is_module(self, obj):
+        return isinstance(obj, (JITFunction, torch.nn.Module))
+
     def is_tensor(self, x, only_native=False):
-        if isinstance(x, (torch.Tensor, JITFunction)):
+        if isinstance(x, torch.Tensor):
             return True
         if only_native:
             return False
@@ -54,6 +58,8 @@ class TorchBackend(Backend):
         return False
 
     def as_tensor(self, x, convert_external=True):
+        if isinstance(x, torch.nn.Module):
+            return x
         if self.is_tensor(x, only_native=convert_external):
             tensor = x
         elif isinstance(x, np.ndarray):
@@ -64,6 +70,7 @@ class TorchBackend(Backend):
             tensor = tensor.to(self.get_default_device().ref)
         elif isinstance(x, (tuple, list)):
             try:
+                x = np.stack(x)
                 tensor = torch.tensor(x, device=self.get_default_device().ref)
             except ValueError:  # there may be Tensors inside the list
                 components = [self.as_tensor(c) for c in x]
@@ -77,13 +84,17 @@ class TorchBackend(Backend):
                 tensor = self.to_float(tensor)
             elif dtype.kind == complex:
                 tensor = self.to_complex(tensor)
+        # --- Move to default device ---
+        if isinstance(tensor, torch.Tensor) and tensor.device != self.get_default_device().ref:
+            tensor = tensor.to(self.get_default_device().ref)
         return tensor
 
     def auto_cast(self, *tensors) -> list:
-        tensors = [t if isinstance(t, (torch.Tensor, numbers.Number, bool)) else self.as_tensor(t, True) for t in tensors]
+        tensors = [t if isinstance(t, (numbers.Number, bool)) else self.as_tensor(t, True) for t in tensors]
         return Backend.auto_cast(self, *tensors)
 
     def is_available(self, tensor) -> bool:
+        # return True
         return torch._C._get_tracing_state() is None  # TODO can we find out whether this tensor specifically is being traced?
 
     def numpy(self, tensor):
@@ -106,14 +117,24 @@ class TorchBackend(Backend):
     def copy(self, tensor, only_mutable=False):
         return torch.clone(tensor)
 
+    def multi_slice(self, tensor, slices: tuple):
+        neg_slices = [i for i, s in enumerate(slices) if isinstance(s, slice) and s.step is not None and s.step < 0]
+        if neg_slices:
+            tensor = torch.flip(tensor, neg_slices)
+        pos_slices = [slice(s.start, s.stop, -s.step) if i in neg_slices else s for i, s in enumerate(slices)]
+        return tensor[tuple(pos_slices)]
+
     sqrt = torch.sqrt
     exp = torch.exp
     sin = torch.sin
+    arcsin = torch.arcsin
     cos = torch.cos
+    arccos = torch.arccos
     tan = torch.tan
     log = torch.log
     log2 = torch.log2
     log10 = torch.log10
+    sigmoid = torch.sigmoid
     isfinite = torch.isfinite
     abs = torch.abs
     sign = torch.sign
@@ -126,7 +147,7 @@ class TorchBackend(Backend):
     einsum = staticmethod(torch.einsum)
 
     def jit_compile(self, f: Callable) -> Callable:
-        return JITFunction(f)
+        return JITFunction(self, f)
 
     def custom_gradient(self, f: Callable, gradient: Callable = None) -> Callable:
         TRACED_F = {}
@@ -142,7 +163,8 @@ class TorchBackend(Backend):
                     f_ = f
 
                     def trace_later():
-                        TRACED_F[current_jit] = torch.jit.trace(f, args)  # nested traces not allowed in PyTorch
+                        PHI_LOGGER.debug(f"Tracing forward pass of '{f.__name__}' which uses a custom gradient")
+                        TRACED_F[current_jit] = torch.jit.trace(f, args, strict=False, check_trace=False)  # nested traces not allowed in PyTorch
                     current_jit.post_trace.append(trace_later)
                 elif CURRENT_JIT_CALLS:
                     jit_f = TRACED_F[CURRENT_JIT_CALLS[-1]]
@@ -165,7 +187,7 @@ class TorchBackend(Backend):
                     # g_ = gradient
                     #
                     # def trace_later():
-                    #     TRACED_F[current_jit] = torch.jit.trace(f, args)  # nested traces not allowed in PyTorch
+                    #     TRACED_F[current_jit] = torch.jit.trace(f, args, check_trace=False)  # nested traces not allowed in PyTorch
                 elif ctx.jit_f:
                     jit_f = ctx.jit_f
                     if ctx.jit_f in TRACED_B:
@@ -175,12 +197,19 @@ class TorchBackend(Backend):
                         # jit-functions cannot return None, so we have to filter out non-required gradients
                         needs_input_grad = ctx.needs_input_grad
                         with ctx.jit_f:
+                            none_indices = []  # jit function cannot return None but gradient returns None to indicate there is no gradient
+
                             def filter_required_grads(*args):
                                 grads = gradient(*args)
                                 filtered = [g for g, need in zip(grads, needs_input_grad) if need]
+                                none_indices.clear()
+                                none_indices.extend([i for i, g in enumerate(filtered) if g is None])
+                                filtered = [g for g in filtered if g is not None]
+                                assert all([isinstance(g, torch.Tensor) for g in filtered]), [type(g) for g in grads]
                                 return filtered
 
                             needed_g = torch.jit.trace(filter_required_grads, tuple([x, y, grad_args]), check_trace=False)
+                            PHI_LOGGER.debug(f"Tracing backward pass of '{f.__name__}' which uses a custom gradient")
 
                             def g_(*args):
                                 with jit_f:
@@ -188,6 +217,8 @@ class TorchBackend(Backend):
                                 assert not jit_f.post_trace
                                 assert isinstance(needed, (tuple, list))
                                 needed = list(needed)
+                                for i in none_indices:
+                                    needed.insert(i, None)
                                 result = [(needed.pop(0) if need else None) for need in needs_input_grad]
                                 return result
 
@@ -207,8 +238,18 @@ class TorchBackend(Backend):
         x, y = self.auto_cast(x, y)
         return x == y
 
-    def random_uniform(self, shape):
-        return torch.rand(size=shape, dtype=to_torch_dtype(self.float_type), device=self.get_default_device().ref)
+    def random_uniform(self, shape, low, high, dtype: DType or None):
+        dtype = dtype or self.float_type
+        if dtype.kind == float:
+            return low + (high - low) * torch.rand(size=shape, dtype=to_torch_dtype(dtype), device=self.get_default_device().ref)
+        elif dtype.kind == complex:
+            real = low.real + (high.real - low.real) * torch.rand(size=shape, dtype=to_torch_dtype(DType(float, dtype.precision)), device=self.get_default_device().ref)
+            imag = low.imag + (high.imag - low.imag) * torch.rand(size=shape, dtype=to_torch_dtype(DType(float, dtype.precision)), device=self.get_default_device().ref)
+            return real + 1j * imag
+        elif dtype.kind == int:
+            return torch.randint(low, high, shape, dtype=to_torch_dtype(dtype))
+        else:
+            raise ValueError(dtype)
 
     def random_normal(self, shape):
         return torch.randn(size=shape, dtype=to_torch_dtype(self.float_type), device=self.get_default_device().ref)
@@ -257,13 +298,15 @@ class TorchBackend(Backend):
         try:
             result = torchf.pad(reordered, pad_width_spatial, mode, value=constant_values)  # supports 3D to 5D (2 + 1D to 3D)
         except RuntimeError as err:
-            warnings.warn(f"PyTorch error {err}")
+            warnings.warn(f"PyTorch error {err}", RuntimeWarning)
             return NotImplemented
         result = undo_transform(result)
         return result
 
-    def grid_sample(self, grid, coordinates, extrapolation='constant'):
+    def grid_sample(self, grid, coordinates, extrapolation: str):
         assert extrapolation in ('undefined', 'zeros', 'boundary', 'periodic', 'symmetric', 'reflect'), extrapolation
+        if get_functional_derivative_order() > 1:
+            return NotImplemented  # PyTorch's grid_sample operator does not define higher-order derivatives
         extrapolation = {'undefined': 'zeros', 'zeros': 'zeros', 'boundary': 'border', 'reflect': 'reflection'}.get(extrapolation, None)
         if extrapolation is None:
             return NotImplemented
@@ -346,22 +389,23 @@ class TorchBackend(Backend):
         return torch.zeros(shape, dtype=to_torch_dtype(dtype or self.float_type), device=self.get_default_device().ref)
 
     def zeros_like(self, tensor):
-        return torch.zeros_like(tensor, device=self.get_default_device().ref)
+        return torch.zeros_like(self.as_tensor(tensor), device=self.get_default_device().ref)
 
     def ones(self, shape, dtype: DType = None):
         return torch.ones(shape, dtype=to_torch_dtype(dtype or self.float_type), device=self.get_default_device().ref)
 
     def ones_like(self, tensor):
-        return torch.ones_like(tensor, device=self.get_default_device().ref)
+        return torch.ones_like(self.as_tensor(tensor), device=self.get_default_device().ref)
 
     def meshgrid(self, *coordinates):
         coordinates = [self.as_tensor(c) for c in coordinates]
-        return torch.meshgrid(coordinates)
+        return torch.meshgrid(*coordinates, indexing='ij')
 
     def linspace(self, start, stop, number):
         return torch.linspace(start, stop, number, dtype=to_torch_dtype(self.float_type), device=self.get_default_device().ref)
 
     def tensordot(self, a, a_axes: tuple or list, b, b_axes: tuple or list):
+        a, b = self.auto_cast(a, b)
         return torch.tensordot(a, b, (a_axes, b_axes))
 
     def matmul(self, A, b):
@@ -381,10 +425,10 @@ class TorchBackend(Backend):
                     values = jit_loop(*values)
                 return values
             else:
-                warnings.warn("Tracing a PyTorch while loop requires an additional tracing pass. You can avoid this by passing a torch.ScriptFunction.")
+                warnings.warn("Tracing a PyTorch while loop requires an additional tracing pass. You can avoid this by passing a torch.ScriptFunction.", RuntimeWarning)
                 raise NotImplementedError()
                 # def trace_later():
-                #     jit_loop = torch.jit.trace(loop)
+                #     jit_loop = torch.jit.trace(loop, check_trace=False)
                 #     @torch.jit.script
                 #     def loop_script(values: Tuple[torch.Tensor], loop_script: Callable):
                 #         while torch.any(values[0]):
@@ -473,6 +517,8 @@ class TorchBackend(Backend):
             return NUMPY.shape(tensor)
 
     def staticshape(self, tensor):
+        if isinstance(tensor, torch.nn.Module):
+            return ()
         if self.is_tensor(tensor, only_native=True):
             return tuple([int(s) for s in tensor.shape])
         else:
@@ -582,7 +628,7 @@ class TorchBackend(Backend):
             multiples = multiples.tolist()
         return self.as_tensor(value).repeat(multiples)
 
-    def sparse_tensor(self, indices, values, shape):
+    def sparse_coo_tensor(self, indices, values, shape):
         indices_ = self.to_int64(indices)
         values_ = self.to_float(values)
         if not self.is_available(values_):
@@ -628,23 +674,28 @@ class TorchBackend(Backend):
         x, residual, iterations, function_evaluations, converged, diverged = torch_sparse_cg_adaptive(lin, y, x0, rtol, atol, max_iter)
         return SolveResult(f"Φ-Flow CG ({'PyTorch*' if self.is_available(y) else 'TorchScript'})", x, residual, iterations, function_evaluations, converged, diverged, "")
 
+    def _prepare_graph_inputs(self, args: tuple, wrt: tuple or list):
+        args = [self.as_tensor(arg, True) if i in wrt else arg for i, arg in enumerate(args)]
+        args = [self.to_float(arg) if self.dtype(arg).kind == int else arg for arg in args]
+        for i, arg in enumerate(args):
+            if self.is_tensor(arg, True) and arg.requires_grad and not arg.is_leaf:
+                arg = torch.clone(arg).detach()
+                arg.requires_grad = i in wrt
+                args[i] = arg
+            elif i in wrt:
+                arg = self.as_tensor(arg, True)
+                arg = arg.detach()  # returns a new tensor in any case
+                arg.requires_grad = True
+                args[i] = arg
+        wrt_args = [arg for i, arg in enumerate(args) if i in wrt]
+        for t in wrt_args:
+            assert t.requires_grad
+        return args, wrt_args
+
     def functional_gradient(self, f, wrt: tuple or list, get_output: bool):
         @wraps(f)
         def eval_grad(*args):
-            args = [self.as_tensor(arg, True) if i in wrt else arg for i, arg in enumerate(args)]
-            for i, arg in enumerate(args):
-                if self.is_tensor(arg, True) and arg.requires_grad and not arg.is_leaf:
-                    arg = torch.clone(arg).detach()
-                    arg.requires_grad = i in wrt
-                    args[i] = arg
-                elif i in wrt:
-                    arg = self.as_tensor(arg, True)
-                    arg = arg.detach()  # returns a new tensor in any case
-                    arg.requires_grad = True
-                    args[i] = arg
-            wrt_args = [arg for i, arg in enumerate(args) if i in wrt]
-            for t in wrt_args:
-                assert t.requires_grad
+            args, wrt_args = self._prepare_graph_inputs(args, wrt)
             output = f(*args)
             loss, aux = (output[0], output[1:]) if isinstance(output, (tuple, list)) else (output, None)
             if loss.ndim > 0:
@@ -653,7 +704,7 @@ class TorchBackend(Backend):
             if get_output:
                 loss = loss.detach()
                 if aux is not None:
-                    aux = [aux_.detach() for aux_ in aux]
+                    aux = [aux_.detach() if isinstance(aux_, torch.Tensor) else aux_ for aux_ in aux]
                     return (loss, *aux, *grads)
                 else:
                     return (loss, *grads)
@@ -661,10 +712,103 @@ class TorchBackend(Backend):
                 return grads
         return eval_grad
 
+    def jacobian(self, f: Callable, wrt: tuple or list, get_output: bool):
+        pass
+
+    def hessian(self, f: Callable, wrt: tuple or list, get_output: bool, get_gradient: bool):
+        # if not get_output and not get_gradient:
+        # @wraps(f)
+        # def eval_hessian(*args):
+        #     batch_size = args[0].shape[0]
+        #     for arg in args:
+        #         assert arg.shape[0] == batch_size, f"All arguments must have a matching batch dimension as their first dimension. Got shapes {[arg.shape for arg in args]}"
+        #
+        #     def f_only_wrt_inputs(*wrt_args_only, reduce_batch=False):
+        #         all_args = list(args)
+        #         for i, arg in zip(wrt, wrt_args_only):
+        #             all_args[i] = arg
+        #         output = f(*all_args)
+        #         loss, aux = (output[0], output[1:]) if isinstance(output, (tuple, list)) else (output, None)
+        #         if reduce_batch:
+        #             if loss.ndim > 0:
+        #                 loss = loss.sum()
+        #         else:
+        #             assert np.prod(loss.shape) == 1, f"Loss (first output of f) must be scalar but has shape {loss.shape}"
+        #             loss = loss.sum()
+        #         return loss
+        #
+        #     wrt_args = tuple([self.as_tensor(arg, True) for i, arg in enumerate(args) if i in wrt])
+        #     result = ()
+        #     if get_output:
+        #         result += f(*args),
+        #     if get_gradient:
+        #         result += torch.autograd.functional.jacobian(lambda *a: f_only_wrt_inputs(*a, reduce_batch=True), wrt_args),
+        #     if hasattr(torch, 'vmap'):
+        #         # single_hessian_f = lambda *args: torch.autograd.functional.hessian(f_only_wrt_inputs, args)
+        #         # multi_hessian_f = torch.vmap
+        #         raise NotImplementedError()
+        #     else:
+        #         hessian = tuple([tuple([[] for _1 in range(len(wrt))]) for _2 in range(len(wrt))])  # n x n matrix of lists
+        #         for b in range(batch_size):
+        #             h = torch.autograd.functional.hessian(f_only_wrt_inputs, tuple([arg[b:b + 1] for arg in wrt_args]))
+        #             for i in range(len(wrt)):
+        #                 for j in range(len(wrt)):
+        #                     fake_batch_dim = args[i].ndim
+        #                     hessian[i][j].append(torch.squeeze(torch.squeeze(h[i][j], fake_batch_dim), 0))
+        #         hessian = [[torch.stack(hessian[i][j]) for j in range(len(wrt))] for i in range(len(wrt))]
+        #         # hessian = torch.stack([torch.autograd.functional.hessian(f_only_wrt_inputs, tuple([arg[b:b+1] for arg in wrt_args])) for b in range(batch_size)])  # manual batch loop
+        #     result += hessian,
+        #     return result
+        # else:
+        @wraps(f)
+        def eval_hessian(*args):
+            args, wrt_args = self._prepare_graph_inputs(args, wrt)
+            output = f(*args)
+            loss, aux = (output[0], output[1:]) if isinstance(output, (tuple, list)) else (output, None)
+            scalar_loss = loss.sum() if loss.ndim > 0 else loss
+            grads = torch.autograd.grad(scalar_loss, wrt_args, create_graph=True, retain_graph=True)  # grad() cannot be called during jit trace
+            hessian = []
+            for grad in grads:
+                hessian.append([[] for _ in grads])
+                for lin_index in range(int(np.prod(grad.shape[1:]))):
+                    multi_index = np.unravel_index(lin_index, grad.shape[1:])
+                    h = torch.autograd.grad(grad[(slice(None),) + multi_index].sum(), wrt_args, allow_unused=True, retain_graph=True)  # grad of every entry in grad
+                    # Warning: This returns incorrect values for certain inputs. Hessian of x^2 returns 0 at x=0 but is correct everywhere else.
+                    # ToDo torch.autograd.functional.hessian does not seem to have this issue. Wait for torch.vmap(), then conditionally switch.
+                    for i, h_ in enumerate(h):
+                        hessian[-1][i].append(h_)
+            for col in hessian:
+                for i, row in enumerate(col):
+                    if len(row) > 1:
+                        col[i] = torch.stack(row, dim=1)
+                    else:
+                        col[i] = row[0]
+                    h_shape = tuple(grads[i].shape) + tuple(grads[i].shape[1:])
+                    col[i] = torch.reshape(col[i], h_shape)
+
+            result = ()
+            if get_output:
+                loss = loss.detach()
+                if aux is not None:
+                    aux = [aux_.detach() if isinstance(aux_, torch.Tensor) else aux_ for aux_ in aux]
+                    result += (loss, *aux),
+                else:
+                    result += loss,
+            if get_gradient:
+                result += tuple([g.detach() for g in grads]),
+            result += hessian,
+            return result
+
+        return eval_hessian
+
     def jit_compile_grad(self, f, wrt: tuple or list, get_output: bool):
         jit = self.jit_compile(f)
         return self.functional_gradient(jit, wrt, get_output)
 
+    def jit_compile_hessian(self, f, wrt: tuple or list, get_output: bool, get_gradient: bool):
+        jit = self.jit_compile(f)
+        return self.hessian(jit, wrt, get_output, get_gradient)
+
     def gradients(self, y, xs: tuple or list, grad_y) -> tuple:
         if self.ndims(y) > 0:
             y = self.sum(y)
@@ -699,7 +843,8 @@ def channels_last(x):
 
 class JITFunction:
 
-    def __init__(self, f):
+    def __init__(self, backend: TorchBackend, f):
+        self.backend = backend
         self.f = f
         self.traced = None
         self.post_trace = []
@@ -707,9 +852,25 @@ class JITFunction:
     def __call__(self, *args, **kwargs):
         if kwargs:
             raise NotImplementedError("kwargs not supported for traced function")
+        args = [self.backend.as_tensor(arg) for arg in args]
         if self.traced is None:
+            native_f = self.f
+
+            class JitModule(torch.nn.Module):
+
+                def __init__(self):
+                    super().__init__()
+                    for submodule in JIT_REGISTERED_MODULES:
+                        self.add_module(str(f"{type(submodule).__name__}_{id(submodule)}"), submodule)
+
+                def forward(self, *args):
+                    PHI_LOGGER.debug(f"Tracing Pytorch jit module for {native_f.__name__}")
+                    return native_f(*args)
+
+            module = JitModule()
             with self:
-                self.traced = torch.jit.trace(self.f, example_inputs=args, check_trace=False)
+                self.traced = torch.jit.trace(module, tuple(args), check_trace=False, strict=False)
+                # self.traced = torch.jit.trace(self.f, tuple(args), check_trace=False, strict=False)
         with self:
             from phi.math.backend import choose_backend
             return choose_backend(self).call(self.traced, *args, name=f"run jit-compiled '{self.f.__name__}'")
@@ -729,6 +890,7 @@ class JITFunction:
 
 
 CURRENT_JIT_CALLS: List[JITFunction] = []
+JIT_REGISTERED_MODULES = []  # Add Networks and other modules to this list before using them in a JIT function!
 
 
 def to_torch_dtype(dtype: DType):
